<?php
namespace systasisgfifscrm;

include_once 'class-affiliate-field.php';

/**
 * Systasis Gravity Forms Infusionsoft Feed Add-On
 *
 * @since     1.0
 * @package   GravityForms
 * @author    Systasis Computer Systems, Inc.
 * @version   1.0.1
 * @copyright Copyright (c) 2017, Systasis Computer Systems
 */

class systasisgfifscrm_GravityFormsInfusionsoftIntegrator extends \GFFeedAddOn
{
    /**
     *
     */
    const REFERRAL_TYPE_COOKIE = 0;
    const REFERRAL_TYPE_PERMANENT = 1;
    const REFERRAL_TYPE_MANUAL = 2;
    const FEED_CONDITIONAL_API_GOALS = 'ConditionalApiGoals';
    const FEED_UNCONDITIONAL_API_GOAL = 'UnConditionalApiGoal';
    const FEED_DUPLICATE_CHECK = 'contactDuplicateCheck';
    const FEED_CHECK_TYPE = 'contactDuplicateCheckType';
    const FEED_LEAD_SOURCE = 'contactLeadSource';
    const FEED_CONTACT_MARKETABLE = 'contactEmailMarketable';
    const FEED_ATTACHMENTS_LIST = 'contactAttachments';
    const ENABLE = 'enable';
    const FIELD_ID = 'field_id';
    const OPERATOR = 'operator';
    const VALUE = 'value';
    const NAME = 'name';
    /**
     * Contains an instance of this class, if available.
     *
     * @since  1.0
     * @access private
     * @var    object $_instance If available, contains an instance of this class.
     */
    private static $_instance = null;
    /**
     * Defines this class' pseudo-type. See javascript,
     */
    protected $_type = 'conditional_api_goal';
    /**
     * Defines the version of the Infusionsoft CRM Feed Add-On.
     *
     * @since  1.0
     * @access protected
     * @var    string $_version Contains the version, defined from infusionsoftcrm.php
     */
    protected $_version = SYSTASISGFIFSCRM_VERSION;
    /**
     * Defines the minimum Gravity Forms version required.
     *
     * @since  1.0
     * @access protected
     * @var    string $_min_gravityforms_version The minimum version required.
     */
    protected $_min_gravityforms_version = '1.9.14.26';
    /**
     * Defines the plugin slug.
     *
     * @since  1.0
     * @access protected
     * @var    string $_slug The slug used for this plugin.
     */
    protected $_slug = 'systasisgfifscrm';
    /**
     * Defines the main plugin file.
     *
     * @since  1.0
     * @access protected
     * @var    string $_path The path to the main plugin file, relative to the plugins folder.
     */
    protected $_path = 'systasisgfifscrm/infusionsoftcrm.php';
    /**
     * Defines the full path to this class file.
     *
     * @since  1.0
     * @access protected
     * @var    string $_full_path The full path.
     */
    protected $_full_path = __FILE__;
    /**
     * Defines the URL where this Feed Add-On can be found.
     *
     * @since  1.0
     * @access protected
     * @var    string The URL of the Feed Add-On.
     */
    protected $_url = 'http://systasis.co';
    /**
     * Defines the title of this Feed Add-On.
     *
     * @since  1.0
     * @access protected
     * @var    string $_title The title of the Feed Add-On.
     */
    protected $_title = 'Gravity Forms Infusionsoft CRM Feed Add-On, from Systasis Computer Systems';
    /**
     * Defines the short title of the Feed Add-On.
     *
     * @since  1.0
     * @access protected
     * @var    string $_short_title The short title.
     */
    protected $_short_title = "Systasis Gravity Forms Infusionsoft Feed Add-On";
    /**
     * Defines if Feed Add-On should use Gravity Forms servers for update data.
     *
     * @since  1.0
     * @access protected
     * @var    bool
     */
    protected $_enable_rg_autoupgrade = false;
    /**
     * Defines the capability needed to access the Feed Add-On settings page.
     *
     * @since  1.0
     * @access protected
     * @var    string $_capabilities_settings_page The capability needed to access the Feed Add-On settings page.
     */
    protected $_capabilities_settings_page = 'systasis_gf_infusionsoftcrm';
    /**
     * Defines the capability needed to access the Feed Add-On form settings page.
     *
     * @since  1.0
     * @access protected
     * @var    string $_capabilities_form_settings The capability needed to access the Feed Add-On form settings page.
     */
    protected $_capabilities_form_settings = 'systasis_gf_infusionsoftcrm';
    /**
     * Defines the capability needed to uninstall the Feed Add-On.
     *
     * @since  1.0
     * @access protected
     * @var    string $_capabilities_uninstall The capability needed to uninstall the Feed Add-On.
     */
    protected $_capabilities_uninstall = 'systasis_gf_infusionsoftcrm_uninstall';
    /**
     * Defines the capabilities needed for the Infusionsoft CRM Feed Add-On
     *
     * @since  1.0
     * @access protected
     * @var    array $_capabilities The capabilities needed for the Feed Add-On
     */
    protected $_capabilities = array('systasis_gf_infusionsoftcrm', 'systasis_gf_infusionsoftcrm_uninstall');
    /**
     * Defines the text translation domain for the Infusionsoft CRM Feed Add-On
     *
     * @since  1.0
     * @access protected
     * @var    string $_domain The text translation domain for the Feed Add-On
     */
    protected $_domain = 'systasisgfifscrm';
    /**
     * Contains an instance of the Infusionsoft CRM API libray, if available.
     *
     * @since  1.0
     * @access protected
     * @var    object $api If available, contains an instance of the Infusionsoft CRM API library.
     */
    protected $api = null;
    /**
     * Defines the transient name used to cache Infusionsoft CRM custom fields.
     *
     * @since  1.0
     * @access protected
     * @var    string $fields_transient_name Transient name used to cache Infusionsoft CRM custom fields.
     */
    protected $fields_transient_name = 'gform_infusionsoftcrm_fields';

    /**
     * Get an instance of this class.
     *
     * @return systasisgfifscrm_GravityFormsInfusionsoftIntegrator
     */
    public static function get_instance()
    {
        if (self::$_instance == null) {
            self::$_instance = new systasisgfifscrm_GravityFormsInfusionsoftIntegrator();
        }

        return self::$_instance;
    }

    /**
     * Autoload the required libraries.
     *
     * @since  1.3.1
     * @access public
     *
     * @uses GFAddOn::is_gravityforms_supported()
     */
    public function pre_init()
    {
        parent::pre_init();

        if ($this->is_gravityforms_supported()) {
        }
    }

    /**
     * Plugin starting point. Handles hooks, loading of language files.
     */
    public function init()
    {
        parent::init();

        \GF_Fields::register(new systasisgfifscrm_AffiliateField());

        // Add a label to the Entry display for this Feed Add-On
        add_filter('gform_entry_meta', function ($entry_meta, $form_id) {
            /*
             * data will be stored with the meta key named contact_id
             * label - entry list will use Id as the column header
             * is_numeric - used when sorting the entry list, indicates whether the data should be treated as numeric when sorting
             * is_default_column - when set to true automatically adds the column to the entry list, without having to edit and add the column for display
             * filter - used on the results pages, the entry list search and export entries page.
             */
            $feeds = \GFAPI::get_feeds(null, $form_id);
            if (!is_a($feeds, "WP_Error")) {
                foreach ($feeds as $feed) {
                    if ($feed['addon_slug'] == $this->_slug) {
                        $entry_meta['contact_id'] = array('label' => 'Contact Id',
                            'is_numeric' => true,
                            'is_default_column' => true,
                            'filter' => array('operators' => array('is', 'isnot', '>', '<')));
                    }
                }
            }

//            $form = $this->form
            return $entry_meta;
        }, 10, 2);
    }

    /**
     *
     * Perform tasks only in the front-end
     *
     * {@inheritDoc}
     * @see \GFAddOn::init_frontend()
     */
    public function init_frontend()
    {
        parent::init_frontend();

        // Attempt to map the GF type to Infusionsoft
        add_filter('gform_addon_field_value', function ($field_value, $form, $entry, $field_id, $slug) {
            if (!rgblank($field_value)) {
                if ($this->_slug == $slug) {
                    $field = \RGFormsModel::get_field($form, $field_id);
                    switch (true) {
                        case (is_a($field, "GF_Field_Date")) :
                            $field_value = \Infusionsoft_ContactService::apiDate($field_value);
                            break;
                    }
                }
            }
            return $field_value;
        }, 10, 5);
    }

    /*
     * Declare minimum software version dependencies.
     *
     * @since  1.0
     * @access public
     *
     * @return array $styles
     */
    public function minimum_requirements()
    {
        $deps = array(
            // Require PHP version 5.6 or higher.
            'php' => array('version' => '5.6'));

        return $deps;
    }

    /**
     * Register needed styles.
     *
     * @since  1.0
     * @access public
     *
     * @return array $styles
     */
    public function styles()
    {
        $styles = array(
            array('handle' => 'gform_infusionsoftcrm_form_settings_css',
                'src' => $this->get_base_url() . '/css/form_settings.css',
                'version' => $this->_version,
                'enqueue' => array(array('admin_page' => array('form_settings')))));

        return array_merge(parent::styles(), $styles);
    }

    /**
     * Register needed scripts.
     *
     * @since  1.0
     * @access public
     *
     * @return array $scripts
     */
    public function scripts()
    {
        $scripts = array(
            array(
                'handle' => 'conditional-api-goal-scripts',
                'src' => $this->get_base_url() . '/js/conditional_api_goal.js',
                'version' => $this->_version,
                'deps' => array('jquery'),
                'enqueue' => array(
                    array(
                        'admin_page' => array('form_settings'),
                        'tab' => $this->_slug
                    )
                )
            ),
            array(
                'handle' => 'gform_placeholder',
                'enqueue' => array(
                    array(
                        'admin_page' => array('form_settings'),
                        'field_types' => array('conditional_api_goal'), // Must match js/conditional_api_goal.js
                    )
                )
            )
        );

        return array_merge(parent::scripts(), $scripts);
    }

    /**
     * Add clear custom fields cache button.
     *
     * @since  1.1
     * @access public
     */
    public function render_uninstall()
    {
        $html = '<div class="hr-divider"></div>';
        $html .= '<h3><span><i class="fa fa-list"></i> ' . esc_html__('Clear Custom Fields Cache', $this->_domain) . '</span></h3>';
        $html .= '<p>' .
            esc_html__('Due to Infusionsoft CRM\'s daily API usage limits, Gravity Forms stores Infusionsoft CRM custom fields data for twelve hours. If you make a change to your custom fields, you might not see it reflected immediately due to this data caching. To manually clear the custom fields cache, click the button below.', $this->_domain) .
            '</p>';
        $html .= '<p><a href="' . add_query_arg('clear_field_cache', 'true') . '" class="button button-primary">' .
            esc_html__('Clear Custom Fields Cache', $this->_domain) . '</a></p>';

        echo $html;

        echo parent::render_uninstall();
    }

    /**
     * Add clear custom fields cache check.
     *
     * @since  1.1
     * @access public
     *
     * @uses GravityFormsInfusionsoftIntegrator::maybe_clear_fields_cache()
     */
    public function plugin_settings_page()
    {
        $this->maybe_clear_fields_cache();

        parent::plugin_settings_page();
    }

    /**
     * Clear the Infusionsoft CRM custom fields cache.
     *
     * @since  1.1
     * @access public
     *
     * @uses GFCommon::add_message()
     */
    public function maybe_clear_fields_cache()
    {

        // If the clear_field_cache parameter isn't set, exit.
        if ('true' !== rgget('clear_field_cache')) {
            return;
        }

        // Clear the cache.
        delete_transient($this->fields_transient_name);

        // Add success message.
        \GFCommon::add_message(esc_html__('Custom fields cache has been cleared.', $this->_domain));
    }

    // # FEED SETTINGS -------------------------------------------------------------------------------------------------

    /**
     * Setup plugin settings fields.
     *
     * @since  1.0
     * @access public
     *
     * @return array
     */
    public function plugin_settings_fields()
    {

        // Prepare plugin description.
        $description = '<p>';
        $description .= sprintf(esc_html__('Infusionsoft CRM is a contact management tool that gives you a 360-degree view of your complete sales cycle and pipeline. Use Gravity Forms to collect customer information and automatically add it to your Infusionsoft CRM account. If you don\'t have a Infusionsoft CRM account, you can %1$ssign up for one here.%2$s', $this->_domain), '<a href="http://www.infusionsoft.com" target="_blank">', '</a>');
        $description .= '</p>';

        return array(array('title' => '', 'description' => $description, 'fields' => array()));
    }

    public function settings_conditional_api_goal($field, $echo = TRUE)
    {
        $checkbox_field = $this->get_conditional_api_goal_checkbox($field);
        $hidden_field = $this->get_conditional_api_goal_hidden_field();
        $goalName_field = $this->get_conditional_api_goal_text_field($field);
        $conditional_logic = $this->get_conditional_api_goal_logic();
        $instructions = $field['instructions'];

        $html = $this->settings_checkbox($checkbox_field, FALSE);
        $html .= $this->settings_hidden($hidden_field, FALSE);
        $html .= $this->settings_text($goalName_field, FALSE);
        $html .= '<div id="conditional_api_goal_conditional_logic_container"><!-- dynamically populated --></div>';
        $html .= '<script type="text/javascript"> (function () {var conditionalApiGoal = new ConditionalApiGoalObj({' .
            'strings: { objectDescription: "' . esc_attr($instructions) . '" },' .
            'goalName: "",' .
            'logicObject: JSON.parse(\'' . $conditional_logic . '\')' .
            '});})() </script>';

        if ($this->field_failed_validation($field)) {
            $html .= $this->get_error_icon($field);
        }

        if ($echo) {
            echo $html;
        }

        return $html;
    }

    public function get_conditional_api_goal_checkbox($field)
    {
        $checkbox_label = isset($field['checkbox_label']) ? $field['checkbox_label'] : esc_html__('Enable Conditional API Goal', $this->_domain);

        $checkbox_field = array(
            'name' => $this->_type . '_conditional_logic',
            'type' => 'checkbox',
            'choices' => array(
                array(
                    'label' => $checkbox_label,
                    'name' => $this->_type . '_conditional_logic',
                ),
            ),
            'onclick' => 'ToggleConditionalLogic( false, "conditional_api_goal" );',
        );

        return $checkbox_field;
    }

    public function get_conditional_api_goal_hidden_field()
    {
        $conditional_logic = $this->get_conditional_api_goal_logic();
        $hidden_field = array(
            'name' => $this->_type . '_logic_object',
            'value' => $conditional_logic,
        );

        return $hidden_field;
    }

    public function get_conditional_api_goal_logic()
    {
        $conditional_logic_object = array();
        $saved_goal = $this->get_setting('conditional_api_goal_logic_object');
        $goal_list = empty($saved_goal) ? [] : $saved_goal;
        $goal = array_shift($goal_list);        

        if (is_array($goal) && count($goal) > 0) {
            $conditional_logic_object = $goal;
        } else {
        	$name = rgars($goal, self::NAME) . '/';
        	$conditional_logic_object = array( 'actionType' => 'show'
        									 , 'logicType' => 'all'
        									 , 'rules' => array( array( 'fieldId' => rgars($goal, $name . self::FIELD_ID)
        								   							  , 'operator' => rgars($goal, $name . self::OPERATOR)
        								   							  , 'value' => rgars($goal, $name . self::VALUE ))));
        }

        $form_id = rgget('id');
        $form = \GFFormsModel::get_form_meta($form_id);
        $conditional_logic = json_encode(\GFFormsModel::trim_conditional_logic_values_from_element($conditional_logic_object, $form));

        return $conditional_logic;
    }

    public function get_conditional_api_goal_text_field($field)
    {
        $text_label = isset($field['text_label']) ? $field['text_label'] : esc_html__('API Goal', $this->_domain);

        $text_field = array(
            'name' => $this->_type . '_name',
            'type' => 'text',
            'value' => ""
        );

        return $text_field;
    }

    public function get_conditional_logic_fields()
    {
        $form = $this->get_current_form();
        $fields = array();

        foreach ($form['fields'] as $field) {
            if ($field->is_conditional_logic_supported()) {
                $inputs = $field->get_entry_inputs();

                if ($inputs) {
                    $choices = array();

                    foreach ($inputs as $input) {
                        if (rgar($input, 'isHidden')) {
                            continue;
                        }
                        $choices[] = array(
                            'value' => $input['id'],
                            'label' => \GFCommon::get_label($field, $input['id'], true)
                        );
                    }

                    if (!empty($choices)) {
                        $fields[] = array('choices' => $choices, 'label' => \GFCommon::get_label($field));
                    }

                } else {
                    $fields[] = array('value' => $field->id, 'label' => \GFCommon::get_label($field));
                }

            }
        }

        return $fields;
    }

    /**
     * Setup fields for feed settings.
     *
     * @since  1.0
     * @access public
     *
     * @uses GFFeedAddOn::get_default_feed_name()
     * @uses GravityFormsInfusionsoftIntegrator::contact_feed_settings_fields()
     *
     * @return array
     */
    public function feed_settings_fields()
    {

        // Prepare base feed settings section.
        $base_fields = array(
            'fields' => array(
                array('name' => 'feedName',
                    'label' => esc_html__('Feed Name', $this->_domain),
                    'type' => 'text',
                    'required' => true,
                    'default_value' => $this->get_default_feed_name(),
                    'tooltip' => '<h6>' . esc_html__('Name', $this->_domain) . '</h6>' .
                        esc_html__('Enter a feed name to uniquely identify this setup.', $this->_domain)),
                array('name' => 'action',
                    'label' => esc_html__('Action', $this->_domain),
                    'required' => true,
                    'type' => 'select',
                    'onchange' => "jQuery(this).parents('form').submit();",
                    'tooltip' => '<h6>' . esc_html__('Action', $this->_domain) . '</h6>' .
                        esc_html__('Choose what will happen when this feed is processed.', $this->_domain),
                    'default_value' => 'contact',
                    'choices' => array(
                        array('label' => esc_html__('Create a New Contact', $this->_domain), 'value' => 'contact')))));

        // Prepare conditional logic settings section.
        $feed_conditional_fields = array('title' => esc_html__('Feed Conditional Logic', $this->_domain),
            // 'dependency' => array('field' => 'action', 'values' => array('contact')),
            'fields' => array(
                array('name' => 'feedCondition',
                    'type' => 'feed_condition',
                    'label' => esc_html__('Conditional Logic', $this->_domain),
                    'checkbox_label' => esc_html__('Enable', $this->_domain),
                    'instructions' => esc_html__('Export to Infusionsoft CRM if', $this->_domain),
                    'tooltip' => '<h6>' . esc_html__('Conditional Logic', $this->_domain) . '</h6>' .
                        esc_html__('When conditional logic is enabled, form submissions will only be exported to Infusionsoft CRM when the condition is met. When disabled, all form submissions will be posted.', $this->_domain))));

        // Prepare conditional API goal logic settings section.
        $conditional_api_goal_fields = array('title' => esc_html__('Conditional API Goal Logic', $this->_domain),
            // 'dependency' => array('field' => 'action', 'values' => array('contact')),
            'fields' => array(array('name' => self::FEED_CONDITIONAL_API_GOALS,
                'type' => $this->_type,
                'label' => esc_html__('Conditional API Goals', $this->_domain),
                'checkbox_label' => esc_html__('Enable', $this->_domain),
                'instructions' => esc_html__('Process this API Goal if', $this->_domain),
                'tooltip' => '<h6>' . esc_html__('Conditional API Goals', $this->_domain) . '</h6>' .
                    esc_html__('Configure conditional API goals. Only certain Contacts will achieve these goals.', $this->_domain))));

        // Get module feed settings sections.
        $contact_fields = $this->contact_feed_settings_fields();
        if (empty($contact_fields)) {
        	\GFCommon::display_admin_message();
        	return null;
        }

        return array($base_fields, $contact_fields, $conditional_api_goal_fields, $feed_conditional_fields);
    }

    /**
     * Setup contact fields for feed settings.
     *
     * @since  1.0
     * @access public
     *
     * @uses GFAddOn::add_field_after()
     * @uses GravityFormsInfusionsoftIntegrator::get_field_map_for_module()
     * @uses GravityFormsInfusionsoftIntegrator::get_file_fields_for_feed_setting()
     * @uses GravityFormsInfusionsoftIntegrator::get_module_field_choices()
     * @uses GravityFormsInfusionsoftIntegrator::get_lead_sources_for_feed_setting()
     *
     * @return array
     */
    public function contact_feed_settings_fields()
    {
    	$contactStandardFields = $this->get_field_map_for_module('Contacts');
    	if (empty($contactStandardFields)) {    		
    		\GFCommon::add_error_message(__('No Infusionsoft fields found!', $this->_domain));
    		$this->log_error(__METHOD__ . '(): No Infusionsoft fields found!');
    		return null;
    	}
        // Prepare contact settings fields.
        $fields = array('title' => esc_html__('Contact Details', $this->_domain),
            // 'dependency' => array('field' => 'action', 'values' => ('contact')),
            'fields' => array(
                array('name' => 'contactStandardFields',
                    'label' => esc_html__('Map Fields', $this->_domain),
                    'type' => 'field_map',
                	'field_map' => $contactStandardFields,
                	'tooltip' => '<h6>' . esc_html__('Map Fields', $this->_domain) . '</h6>' .
                		esc_html__('Select which Gravity Form fields pair with their respective Infusionsoft CRM fields.', $this->_domain)),
            	array('name' => 'contactCustomFields',
                    'label' => null,
                    'type' => 'dynamic_field_map',
                    'disable_custom' => true,
                    'exclude_field_types' => 'time',
                    'field_map' => $this->get_field_map_for_module('Contacts', 'dynamic')),
                array('name' => self::FEED_LEAD_SOURCE,
                    'label' => esc_html__('Lead Source', $this->_domain),
                    'type' => 'select_custom',
                    'choices' => $this->get_lead_sources_for_feed_setting(),
                    'tooltip' => '<h6>' . esc_html__('Lead Source', $this->_domain) . '</h6>' .
                        esc_html__('Select a pre-defined lead source, or select "Add Custom Lead Source", then enter a feed-specific value.', $this->_domain)),
                array('name' => self::FEED_UNCONDITIONAL_API_GOAL,
                    'label' => esc_html__('Unconditional API Goal', $this->_domain),
                    'type' => 'text',
                    'tooltip' => '<h6>' . esc_html__('Unconditional API Goal. Every Contact will achieve this goal.', $this->_domain) . '</h6>' . esc_html('API Goal')),
                array('name' => self::FEED_CONTACT_MARKETABLE,
                    'label' => esc_html__('Mark contact marketable', $this->_domain),
                    'type' => 'checkbox',
                    'choices' => array(
                        array('name' => self::FEED_CONTACT_MARKETABLE, 'label' => 'Yes', 'default_value' => 1, 'value' => 1)),
                    'tooltip' => '<h6>' . esc_html__('Mark contact marketable', $this->_domain) . '</h6>' .
                        esc_html__('Contact consents to receive marketing from you. Does not confirm the email. Consider implementing a campaign to send email confirmation.', $this->_domain)),
                array('name' => self::FEED_DUPLICATE_CHECK,
                    'label' => esc_html__('Avoid duplicate contacts', $this->_domain),
                    'type' => 'checkbox_and_select',
                    'checkbox' => array('name' => self::FEED_DUPLICATE_CHECK,
                        'label' => esc_html__('Yes', $this->_domain)),
                    'select' => array('name' => self::FEED_CHECK_TYPE,
                        'choices' => array(array('label' => 'and dup check using email', 'value' => 'Email'),
                            array('label' => 'and dup check using email, name', 'value' => 'EmailAndName'),
                            array('label' => 'and dup check using email, name, company', 'value' => 'EmailAndNameAndCompany'))),
                    'tooltip' => '<h6>' . esc_html__('Avoid duplicate contacts', $this->_domain) . '</h6>' .
                        esc_html__('Yes: Attempt to update existing contact based on value(s) to find, otherwise add.</br>No: Always add the contact, possibly creating duplicates.', $this->_domain))));

        // Get file field choices.
        $file_choices = $this->get_file_fields_for_feed_setting('contact');

        // Add Attachments settings field if file field choices exist.
        if (!empty($file_choices)) {

            // Prepare Attachments settings field.
            $file_field = array('name' => self::FEED_ATTACHMENTS_LIST,
                'type' => 'checkbox',
                'label' => esc_html__('Attachments', $this->_domain),
                'choices' => $file_choices,
                'tooltip' => '<h6>' . esc_html__('Attachments', $this->_domain) . '</h6>' .
                    esc_html__('Infusionsoft CRM has a maximum file size of 10MB. Any file larger than this will not be uploaded.', $this->_domain));

            // Add settings field.
            $fields = $this->add_field_after('contactLeadSource', $file_field, $fields);
        }

        return $fields;
    }

    /**
     * Get field map fields for a Infusionsoft CRM module.
     *
     * @since  1.0
     * @access public
     *
     * @param string $module Module name.
     * @param string $field_map_type Type of field map: standard or dynamic. Defaults to standard.
     *
     * @return array $field_map
     */
    public function get_field_map_for_module($module, $field_map_type = 'standard')
    {

        // Initialize field map.
        $field_map = array();

        // Define standard field labels.
        $standard_fields = array('Company', 'Email', 'FirstName', 'LastName');

        // Get fields for module.
        $fields = $this->get_module_fields($module);

        // Sort module fields in alphabetical order.
        usort($fields, array($this, 'sort_module_fields_by_label'));

        // Loop through module fields.
        foreach ($fields as $field) {
            // If this is a non-supported field type, skip it.
            if (in_array($field['type'], array('Date/Time', 'User', 'Drilldown', 'Social Security Number'))) {
                continue;
            }

            // If this is a standard field map and the field is not a standard field or is not required, skip it.
            if ('standard' === $field_map_type && !$field['required'] && !in_array($field['label'], $standard_fields)) {
                continue;
            }

            // If this is a dynamic field map and the field matches a standard field or is required, skip it.
            if ('dynamic' === $field_map_type && ($field['required'] || in_array($field['label'], $standard_fields))) {
                continue;
            }

            // Get Gravity Forms field type.
            switch ($field['type']) {
                case 'Year' :
                case 'Date' :
                    $field_type = 'date';
                    break;

                case 'Email' :
                    $field_type = array('email', 'hidden');
                    break;

                case 'Phone' :
                    $field_type = 'phone';
                    break;

                default :
                    $field_type = null;
                    break;
            }

            // Add field to field map.
            $field_map[] = array('name' => $field['name'],
                'label' => $field['label'],
                'value' => $field['name'],
                'required' => $field['required'],
                'field_type' => $field_type);
        }

        return $field_map;
    }

    /**
     * Get fields for a Infusionsoft CRM module.
     *
     * @since  1.0
     * @access public
     *
     * @param string $module Module to get fields for. Defaults to all modules.
     *
     * @uses GravityFormsInfusionsoftIntegrator::update_cached_fields()
     *
     * @return array
     */
    public function get_module_fields($module = null)
    {
        // If module fields are not cached, retrieve them.
        if (false === ($fields = get_transient($this->fields_transient_name))) {
            $fields = $this->update_cached_fields();
        }

        $fields = json_decode($fields, true);

        return rgar($fields, $module) ? rgar($fields, $module) : $fields;
    }

    /**
     * Update the cached fields for all the needed modules.
     *
     * @since  1.0
     * @access public
     *
     * @uses GravityFormsInfusionsoftIntegrator::get_fields()
     *
     * @return string $fields JSON encoded string of all module fields.
     */
    public function update_cached_fields()
    {
        /*
         * Key - Infusionsoft type id
         * Value - array(Infusionsoft type name, has options?)
         */
        $typeMaps = [1 => ['Phone Number', false],
            2 => ['Social Security Number', false],
            3 => ['Currency', false],
            4 => ['Percent', false],
            5 => ['State', false],
            6 => ['Yes/No', false],
            7 => ['Year', false],
            8 => ['Month', false],
            9 => ['Day of Week', false],
            10 => ['Name', false],
            11 => ['Decimal Number', false],
            12 => ['Whole Number', false],
            13 => ['Date', false],
            14 => ['Date/Time', false],
            15 => ['Text', false],
            16 => ['Text Area', false],
            17 => ['List Box', true],
            18 => ['Website', false],
            19 => ['Email', false],
            20 => ['Radio', true],
            21 => ['Dropdown', true],
            22 => ['User', true],
            23 => ['Drilldown', true]];

        // If API instance is not initialized, exit.
        if (!$this->initialize_api()) {
            $this->log_error(__METHOD__ . '(): Unable to update fields because API is not initialized.');

            return '';
        }

        try {
            // Get module fields.
            $contact = new \Infusionsoft_Contact();
            $customFields = \Infusionsoft_DataService::getCustomFields($contact);
            $contact->addCustomFields($customFields);
            $contactFields = $contact->getFields();
        } catch (\Exception $e) {
            $this->log_error(__METHOD__ . '(): Unable to update fields. ' . $e->getMessage());

            return '';
        }

        // Initialize fields array.
        $fields = array();

        // Loop through modules.
        foreach ($contactFields as $contactField) {
            // Prepare field details.
            $typeMap = is_array($contactField) ? rgar($typeMaps, $contactField[0]->DataType) : "";
            $field = array('custom_field' => is_array($contactField),
                'label' => is_array($contactField) ? $contactField[0]->Label : $contactField,
                'name' => is_array($contactField) ? $contactField[0]->Name : $contactField,
                'required' => false,
                'type' => is_array($contactField) ? (!empty($typeMap) ? $typeMap[0] : "") : "");

            // Store field choices, if set.
            if (!empty($typeMap)) {
                if ($typeMap[1]) {
                    $field['choices'] = $contactField[0]->Values;
                }
            }

            // Add field to array.
            $fields[] = $field;
        }

        // Convert fields array to JSON string.
        $fields = json_encode($fields);

        // Store fields.
        set_transient($this->fields_transient_name, $fields, 60 * 60 * 12);

        return $fields;
    }

    // # FEED PROCESSING -----------------------------------------------------------------------------------------------

    /**
     * Initializes the Infusionsoft CRM API.
     *
     * @since  1.0
     * @access public
     *
     * @uses Infusionsoft_AppPool::addApp()
     * @uses Infusionsoft_AppPool::getApp()
     * @uses Infusionsoft_App:enableDebug()
     *
     * @return bool API initialization state.
     */
    public function initialize_api()
    {

        // If the API is already initialized, return true.
        if (!is_null($this->api)) {
            return true;
        }

        /**
         * Get the IFS API credentials
         */
        try {
        	$this->api = \Infusionsoft_AppPool::addApp(new \Infusionsoft_App(get_option('infusionsoft_sdk_app_name') . '.infusionsoft.com',
                get_option('infusionsoft_sdk_api_key')));
            if (! \Infusionsoft_DataService::ping()) {
            	throw new \Exception('Infusionsoft connection failed, check App name and API key values on the Novak SDK settings page');
            }
        } catch (\Exception $e) {
        	\GFCommon::add_error_message($e->getMessage());
        	$this->log_error(__METHOD__ . '(): ' . $e->getMessage());
            $this->api = null;
        }

        // @var Infusionsoft_App
        if (! empty($this->api)) {
        	$this->api->enableDebug();
        }
        
        return (!is_null($this->api));
    }

    /**
     * Get Infusionsoft CRM lead sources.
     *
     * @since  1.0
     * @access public
     *
     * @uses GFAddOn::log_error()
     *
     * @return array $lead_sources
     */
    public function get_lead_sources_for_feed_setting()
    {

        // Initialize choices array.
        $choices = array(array('label' => esc_html__('-None-', $this->_domain), 'value' => ''));

        // If API instance is not initialized, return choices.
        if (!$this->initialize_api()) {
            $this->log_error(__METHOD__ . '(): Unable to get lead sources because API is not initialized.');
            return $choices;
        }

        try {

            // Get Infusionsoft lead sources.
            $lead_sources = \Infusionsoft_DataService::queryWithOrderBy(new \Infusionsoft_LeadSource(), array('Status' => 'Active'), 'Name');
        } catch (\Exception $e) {

            // Log that lead sources could not be retrieved.
            $this->log_error(__METHOD__ . '(): Unable to get lead sources; ' . $e->getMessage());

            return $choices;
        }

        // If Infusionsoft CRM lead sources exist, add them as choices.
        if (!empty($lead_sources)) {

            // Loop through Infusionsoft lead sources.
            foreach ($lead_sources as $lead_source) {

                // Add lead source as choice.
                $choices[] = array('label' => esc_html($lead_source->Name), 'value' => esc_attr($lead_source->Name));
            }
        }

        return $choices;
    }

    /**
     * Get form file fields for feed field settings.
     *
     * @since  1.0
     * @access public
     *
     * @param  string $module Module to prepare file fields for. Defaults to contact.
     *
     * @uses GFAddOn::get_current_form()
     * @uses GFAPI::get_fields_by_type()
     *
     * @return array
     */
    public function get_file_fields_for_feed_setting($module = 'contact')
    {

        // Initialize choices array.
        $choices = array();

        // Get the form.
        $form = $this->get_current_form();

        // Get file fields.
        $file_fields = \GFAPI::get_fields_by_type($form, array('fileupload'), true);

        // If file fields exist, prepare them as choices.
        if (!empty($file_fields)) {

            // Loop through file fields.
            foreach ($file_fields as $field) {

                // Add file field as choice.
                $choices[] = array('name' => $module . 'Attachments[' . $field->id . ']', 'label' => $field->label, 'default_value' => 0);
            }
        }

        return $choices;
    }

    /**
     * Insert settings field after another field.
     * (Forked to allow for passing a single settings section.)
     *
     * @since  1.3.1
     * @access public
     *
     * @param string $name Field name to insert settings field after.
     * @param array $fields Settings field.
     * @param array $settings Settings section to add field to.
     *
     * @return array
     */
    public function add_field_after($name, $fields, $settings)
    {

        // Move settings into another array.
        $settings = array($settings);

        // Add field.
        $settings = parent::add_field_after($name, $fields, $settings);

        // Return the first settings section.
        return $settings[0];
    }

    public function maybe_save_feed_settings($feed_id, $form_id)
    {
        /*
         * Validate goal name
         */
        $settings = $this->get_posted_settings();
        $settings = $this->trim_conditional_logic_vales($settings, $form_id);
        $conditional_api_goal_enabled = rgar($settings,  $this->_type . '_conditional_logic');

        if ($conditional_api_goal_enabled === '1' && $this->is_save_postback() === true) {

            /*
             * If conditional api goal name is set, save feed setting as usual
             *
             * Otherwise, show error
             */
            if (!(empty(rgar($settings, $this->_type . '_name')))) {
                return parent::maybe_save_feed_settings($feed_id, $form_id);
            } else {
                \GFCommon::add_error_message('You need to set conditional API goal name first.');
            }
        } else {
            return parent::maybe_save_feed_settings($feed_id, $form_id);
        }
    }

    /**
     * Get choices for a specifc Infusionsoft CRM module field formatted for field settings.
     *
     * @since  1.0
     * @access public
     *
     * @param string $module
     * @param string $field_name
     *
     * @uses GravityFormsInfusionsoftIntegrator::get_module_field()
     *
     * @return array
     */
    public function get_module_field_choices($module, $field_name)
    {
        // Initialize choices array.
        $choices = array();

        // Get module field for field name.
        $field = $this->get_module_field($module, $field_name);

        // If no field choices exist, return choices.
        if (empty($field['choices'])) {
            return $choices;
        }

        // Loop through field choices.
        foreach ($field['choices'] as $choice) {

            // If choice is an array, use content as choice.
            if (is_array($choice) && rgar($choice, 'content')) {
                $choice = $choice['content'];
            }

            // Add field choice as choice.
            $choices[] = array('label' => esc_html($choice), 'value' => esc_attr($choice));
        }

        return $choices;
    }

    /**
     * Get field from a Infusionsoft CRM module.
     *
     * @since  1.0
     * @access public
     *
     * @param string $module Module to get field from.
     * @param string $field_name Field name to retrieve.
     *
     * @uses GravityFormsInfusionsoftIntegrator::get_module_fields()
     *
     * @return array
     */
    public function get_module_field($module, $field_name)
    {

        // Get fields for module.
        $module_fields = $this->get_module_fields($module);

        // Loop through module fields.
        foreach ($module_fields as $module_field) {

            // If field label matches the field name, return field.
            if (rgar($module_field, 'label') === $field_name) {
                return $module_field;
            }
        }

        return array();
    }

    /**
     * Insert settings field before another field.
     * (Forked to allow for passing a single settings section.)
     *
     * @since  1.3.1
     * @access public
     *
     * @param string $name Field name to insert settings field after.
     * @param array $fields Settings field.
     * @param array $settings Settings section to add field to.
     *
     * @return array
     */
    public function add_field_before($name, $fields, $settings)
    {

        // Move settings into another array.
        $settings = array($settings);

        // Add field.
        $settings = parent::add_field_before($name, $fields, $settings);

        // Return the first settings section.
        return $settings[0];
    }    	

    // # HELPER FUNCTIONS ----------------------------------------------------------------------------------------------

    /**
     * Set feed creation control.
     *
     * @since  1.0
     * @access public
     *
     * @return bool
     */
    public function can_create_feed()
    {
        return $this->initialize_api();
    }
    
    public function configure_addon_message() {
    	return "<div class=\"error\"><p><strong>Please set your Infusionsoft SDK app name and API key on the <a href=\"" . admin_url( 'options-general.php?page=infusionsoft-sdk/infusionsoft-sdk.php' ) . "\">settings page.</a></strong></p></div>";
    }

    /**
     * Enable feed duplication.
     *
     * @since  1.1.7
     * @access public
     *
     * @param int $id Feed ID requesting duplication.
     *
     * @return bool
     */
    public function can_duplicate_feed($id)
    {
        return true;
    }

    /**
     * Setup columns for feed list table.
     *
     * @since  1.0
     * @access public
     *
     * @return array
     */
    public function feed_list_columns()
    {
        return array('feedName' => esc_html__('Name', $this->_domain), 'action' => esc_html__('Action', $this->_domain));
    }

    /**
     * Get value for action feed list column.
     *
     * @since  1.0
     * @access public
     *
     * @param  array $feed Feed for current table row.
     *
     * @return string
     */
    public function get_column_value_action($feed)
    {

        // Display contact action string.
        if (rgars($feed, 'meta/action') == 'contact') {
            return esc_html__('Create a New Contact', $this->_domain);
        }

        return esc_html__('No Action', $this->_domain);
    }

    /**
     * Process the Infusionsoft CRM feed.
     *
     * @since  1.0
     * @access public
     *
     * @param  array $feed Feed object.
     * @param  array $entry Entry object.
     * @param  array $form Form object.
     *
     * @uses GFAddOn::log_debug()
     * @uses GFAddOn::log_error()
     * @uses GravityFormsInfusionsoftIntegrator::create_contact()
     * @uses GravityFormsInfusionsoftIntegrator::initialize_api()
     * @uses GravityFormsInfusionsoftIntegrator::upload_attachments()
     */
    public function process_feed($feed, $entry, $form)
    {
        // If API instance is not initialized, exit.
        if (!$this->initialize_api()) {

            // Log that we cannot process the feed.
            $this->add_feed_error(esc_html__('Feed was not processed because API was not initialized.', $this->_domain), $feed, $entry, $form);

            return;
        }

        // Create contact.
        switch (rgars($feed, 'meta/action')) {
            case 'contact' :
                // Get contact ID.
                $contact_id = $this->create_contact($feed, $entry, $form);

                if (rgblank($contact_id)) {
                    return;
                }

                // Upload attachments
                $this->upload_attachments($contact_id, $feed, $entry, $form);

                // Affiliate referral tracking
                $affiliate_code = $this->track_affiliate_referral($contact_id, $feed, $entry, $form);

                /**
                 * There will be two different API goal metadata fields. One conditional, the other unconditional.
                 * The API goal name is required when the conditional API goal is enabled
                 */

                /*
                 * Get and process any conditional API goals
                 *
                 * Get all conditional goals
     			 * @since  2.0
                 */
                $r0 = rgars($feed, 'meta/' . $this->_type . '_logic_object');
                $goals[] = array( 'enabled' => rgars($feed, 'meta/' . $this->_type . '_conditional_logic')
                                , 'conditionalLogic' => array_shift($r0)
                                , 'name' => rgars($feed, 'meta/' . $this->_type . '_name'));

                /*
                * Process the conditional goal list
                */
                foreach ($goals as $goal) {
                    if ($this->is_conditional_api_goal_condition_met($form, $entry, $goal)) {
                        $api_goal_name = rgars($goal, self::NAME);
                        $this->achieve_goal($feed, $entry, $form, $api_goal_name, $contact_id);
                    } else {
                        $this->log_debug(__METHOD__ . '(): condition not met for contact id: ' . $contact_id);
                    }
                }
                
                /*
                 * Get and process any unconditional API goal
                 */
                $unconditionalApiGoal = rgars($feed, 'meta/' . self::FEED_UNCONDITIONAL_API_GOAL);

                if (!empty($unconditionalApiGoal)) {
                    $this->achieve_goal($feed, $entry, $form, $unconditionalApiGoal, $contact_id);
                } else {
                    $this->log_debug(__METHOD__ . '(): No unconditional goal for contact id: ' . $contact_id);
                }

                break;
        }
    }

    /**
     * Create a new Infusionsoft contact from a feed.
     *
     * @since  1.0
     * @access public
     *
     * @param array $feed Feed object.
     * @param array $entry Entry object.
     * @param array $form Form object.
     *
     * @uses GFAddOn::get_field_map_fields()
     * @uses GFAddOn::get_field_value()
     * @uses GFAddOn::log_debug()
     * @uses GFAddOn::log_error()
     * @uses GFCommon::replace_variables()
     * @uses GFFeedAddOn::add_feed_error()
     * @uses GravityFormsInfusionsoftIntegrator::get_dynamic_field_map_fields()
     * @uses Infusionsoft_ContactService::addWithDupCheck()
     * @uses Infusionsoft_ContactService::add()
     *
     * @return int|null $contact_id
     */
    public function create_contact($feed, $entry, $form)
    {

        // Initialize contact object.
        $r0 = rgars($feed, 'meta/' . self::FEED_LEAD_SOURCE);
        $contact["Leadsource"] = (false == stristr($r0, 'gf_custom')) ? $r0 : rgars($feed, 'meta/' . self::FEED_LEAD_SOURCE . '_custom');
        
        // Get standard and custom fields.
        $standard_fields = $this->get_field_map_fields($feed, 'contactStandardFields');
        $custom_fields = $this->get_dynamic_field_map_fields($feed, 'contactCustomFields');

        // Merge standard and custom fields arrays.
        $mapped_fields = array_merge($standard_fields, $custom_fields);

        // Loop through mapped fields.
        foreach ($mapped_fields as $field_name => $field_id) {
            // Get field value.
            $field_value = $this->get_field_value($form, $entry, $field_id);

            // If field value is empty, skip it.
            if (rgblank($field_value)) {
                continue;
            }

            // Add mapped field to contact object.
            $contact[$field_name] = $field_value;
        }

        /**
         * Modify the contact arguments before they are sent to Infusionsoft CRM.
         *
         * @since 1.0
         *
         * @param int $form ['id'] The form id.
         * @param array $contact The Infusionsoft contact.
         * @param array $feed Feed object.
         * @param array $entry Entry object.
         * @param array $form Form object.
         */
        $contact = gf_apply_filters('gform_infusionsoftcrm_contact', $form['id'], $contact, $feed, $entry, $form);

        // Create new contact
        $contact_id = null;
        try {
            if (1 == rgars($feed, 'meta/' . self::FEED_DUPLICATE_CHECK)) {
                $duplicateCheckType = rgars($feed, 'meta/' . self::FEED_CHECK_TYPE);
                $contact_id = \Infusionsoft_ContactService::addWithDupCheck($contact, $duplicateCheckType);
            } else {
                $contact_id = \Infusionsoft_ContactService::add($contact);
            }

            $this->log_debug(__METHOD__ . '(): Contact #' . $contact_id . ' created.');
        } catch (\Exception $e) {
            // Log that contact could not be created.
            $this->add_feed_error('Could not create contact; ' . esc_html($e->getMessage()), $feed, $entry, $form);
            $contact_id = null;
        }

        gform_update_meta($entry['id'], 'contact_id', $contact_id, $form['id']);

        // Maybe mark email as marketable
        if (1 == rgars($feed, 'meta/' . self::FEED_CONTACT_MARKETABLE))
        {
            try {
            	if (isset($contact["Email"])) {
                	$this->optin($contact["Email"]);
            	}
            } catch (\Exception $e) {
                $this->add_feed_error(esc_html($e->getMessage()), $feed, $entry, $form);
            }
        }

        return $contact_id;
    }

    public function optin($email = null) {
        /*
         * Attempt email opt-in
         */
        if (isset($email)) {
            try {
                $title = $this->_short_title;
                \Infusionsoft_EmailService::optIn($email, $title);
                $this->log_debug(__METHOD__ . '(): ' . $email . ' flagged marketable.');
            } catch (\Exception $e) {
                throw new \Exception("Could not flag email marketable");
            }
        }
    }

    /**
     * Upload attachments from a feed.
     *
     * @since  1.0
     * @access public
     *
     * @param int $contact_id Contact ID to add attachment.
     * @param array $feed Feed object.
     * @param array $entry Entry object.
     * @param array $form Form object.
     *
     * @uses GFAddOn::get_field_value()
     * @uses GFAddOn::is_json()
     * @uses GFAddOn::log_debug()
     * @uses GFAddOn::log_error()
     * @uses Infusionsoft_FileService::uploadFile()
     */
    public function upload_attachments($contact_id, $feed, $entry, $form)
    {

        $attachment_list = rgars($feed, 'meta/' . self::FEED_ATTACHMENTS_LIST);

        // If no file upload fields are selected as attachments, exit.
        if (0 == count($attachment_list) || !is_array($attachment_list)) {
            return;
        }

        // Initialize array to store file upload fields.
        $file_fields = array();

        // Loop through attachments settings field choices.
        foreach ($attachment_list as $field_id => $value) {

            // If this field is enabled for attachments, add it to the file upload fields array.
            if ('1' == $value) {
                $file_fields[] = $field_id;
            }
        }

        // If no file upload fields are defined, exit.
        if (empty($file_fields)) {
            return;
        }

        // Loop through file upload fields.
        foreach ($file_fields as $file_field) {

            // Get files for field.
            $files = $this->get_field_value($form, $entry, $file_field);

            // If no files were uploaded for this field, skip it.
            if (empty($files)) {
                continue;
            }

            // Convert files value to array.
            $files = $this->is_json($files) ? json_decode($files, true) : explode(' , ', $files);

            // Loop through the files.
            foreach ($files as $i => &$file) {

                // Convert file URL to local path.
                $file_path = str_replace(WP_CONTENT_URL, WP_CONTENT_DIR, $file);

                // If the file is larger than the maximum allowed by Infusionsoft CRM, skip it.
                if (filesize($file_path) > 10000000) {
                    $this->log_error(__METHOD__ . '(): Unable to upload file "' . basename($file_path) . '" because it is larger than 10MB.');
                    continue;
                }

                // Upload file.
                try {
                    \Infusionsoft_FileService::uploadFile($contact_id, basename($file_path), base64_encode(file_get_contents($file_path)));
                    $this->log_debug(__METHOD__ . '(): File "' . basename($file_path) . '" has been uploaded to contact #' . $contact_id . '.');
                } catch (\Exception $e) {
                    $this->log_error(__METHOD__ . '(): File "' . basename($file_path) . '" could not be uploaded; ' . $e->getMessage());
                }
            }
        }
    }

    /**
     *
     * @abstract Evaluate Conditional Logic for an API Goal block
     * @since  2.0
     *
     * @param array GF Feed object
     * @param array GF Form object
     * @param array API Goal block
     */
    public function is_conditional_api_goal_condition_met($form, $entry, $goal = null)
    {
        if (false == rgars($goal, "enabled")) {
            // The goal is not enabled so we handle it as if the rules are not met.
            return FALSE;
        }

        return \GFCommon::evaluate_conditional_logic($goal['conditionalLogic'], $form, $entry);
    }

    /**
     * Start a contact in an Infusionsoft campaign
     *
     * @since  1.0
     * @access public
     *
     * @param array $feed Feed object.
     * @param array $entry Entry object.
     * @param array $form Form object.
     * @param string $api_goal Infusionsoft API goal.
     * @param integer $contact_id Infusionsoft contact id.
     *
     * @uses GFAddOn::add_feed_error()
     * @uses GFAddOn:log_debug()
     * @uses Infusionsoft_FunnelService::achieveGoal()
     */
    public function achieve_goal($feed, $entry, $form, $api_goal = "", $contact_id = 0)
    {
        $integration = get_option('infusionsoft_sdk_app_name');

        if (!empty($api_goal)) {
            try {
                $result = \Infusionsoft_FunnelService::achieveGoal($integration, $api_goal, $contact_id);
                if (0 < strlen($result[0]["msg"])) {
                    throw new \Exception($result[0]["msg"]);
                }
                $this->log_debug(__METHOD__ . '(): API Goal ' . $api_goal . ' set for contact id:' . $contact_id);
            } catch (\Exception $e) {
                // Log that contact could not be started in campaign.
                $this->add_feed_error('Could not start contact in campaign; ' . esc_html($e->getMessage()), $feed, $entry, $form);
            }
        } else {
            $this->log_debug(__METHOD__ . '(): No API goal for contact id: ' . $contact_id);
        }
    }

    /**
     * Create an affiliate tracking record
     *
     * @param integer $contact_id
     * @param array $feed Feed object.
     * @param array $entry Entry object.
     * @param array $form Form object.
     *
     * @uses AffiliateField::getAffiliateId()
     * @uses Infusionsoft_DataService::apiDateFormat
     *
     * @return NULL|string|mixed
     */
    public function track_affiliate_referral($contact_id, $feed, $entry, $form)
    {
        $affiliate_id = $affiliate_code = null;

        foreach ($form['fields'] as $field) {
            if (is_a($field, 'AffiliateField')) {
                $affiliate_code = rgar($entry, $field->id);
            }
        }

        if (!empty($affiliate_code)) {
            try {
                $affiliate_id = systasisgfifscrm_AffiliateField::getAffiliateId($affiliate_code);
            } catch (\Exception $e) {
                $this->add_feed_error('Could not track affiliate referral; ' . esc_html($e->getMessage()), $feed, $entry, $form);
            }

            if (!empty($affiliate_id)) {
                $referral = new \Infusionsoft_Referral();
                $referral->ContactId = $contact_id;
                $referral->AffiliateId = $affiliate_id;
                $referral->DateSet = date(\Infusionsoft_DataService::apiDateFormat);
                $referral->Type = self::REFERRAL_TYPE_MANUAL;
                $referral->IPAddress = rgar($_SERVER, "REMOTE_ADDR");
                $referral->Source = $this->_short_title;

                try {
                    $referral->save();
                    $this->log_debug(__METHOD__ . '(): Affiliate referral for ' . $affiliate_code . ' created.');
                } catch (\Exception $e) {
                    $this->add_feed_error('Could not track affiliate referral; ' . esc_html($e->getMessage()), $feed, $entry, $form);
                }
            }
        }
        return $affiliate_code;
    }

    /**
     * Sort module fields alphabeically by label.
     *
     * @since  1.0
     * @access public
     *
     * @param  array $a First array item.
     * @param  array $b Second array item.
     *
     * @return int
     */
    public function sort_module_fields_by_label($a, $b)
    {
        return strcmp($a['label'], $b['label']);
    }
}