<?php
namespace systasisgfifscrm;

/*
 Plugin Name: Systasis GF Infusionsoft Feed
 Plugin URI: http://systasis.co
 Description: Integrates Gravity Forms with Infusionsoft CRM, allowing form submissions to be automatically sent to your Infusionsoft CRM account.
 Version: 2.2.3
 Author: Systasis Computer Systems, Inc.
 Text Domain: systasisgfifscrm

 ------------------------------------------------------------------------
 Copyright 2017, Systasis Computer Systems, Inc.
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see http://www.gnu.org/licenses. 
 */
define('SYSTASISGFIFSCRM_VERSION', '2.2.3');

class systasisgfifscrm_GravityFormsInfusionsoftIntegrator_Bootstrap
{

    /**
     * If the Feed Add-On Framework exists, InfusionsoftGravityForms_Plugin Add-On is loaded.
     *
     * @since  1.0
     * @access public
     */
    public static function load()
    {
		if (! method_exists('GFForms', 'include_feed_addon_framework')) {return;}

        require_once('class-gravityforms-infusionsoft-integrator.php');

        \GFAddOn::register('systasisgfifscrm\systasisgfifscrm_GravityFormsInfusionsoftIntegrator');
    }
}

// If Gravity Forms is loaded, bootstrap the InfusionsoftGravityForms_Plugin Add-On.
// Read more here: https://docs.gravityforms.com/gform_loaded/
add_action('gform_loaded', array('systasisgfifscrm\systasisgfifscrm_GravityFormsInfusionsoftIntegrator_Bootstrap', 'load'), 5);

/*
 * Check for requirements when plugin is activated
 */
function systasisgfifscrm_check_for_requirements()
{
    if (!is_plugin_active('gravityforms/gravityforms.php') || !is_plugin_active('infusionsoft-sdk/infusionsoft-sdk.php')) {
        deactivate_plugins(plugin_basename('systasisgfifscrm\infusionsoftcrm.php'));
        wp_die('This plugin requires GravityForms and Infusionsoft SDK. Please ensure you have them installed.');
        return;
    }    
}
register_activation_hook(__FILE__, 'systasisgfifscrm\systasisgfifscrm_check_for_requirements');

/*
 * Function required_plugins_notices
 *
 * It checks for all required plugins on current Wordpress installation and show notifications
 */
function systasisgfifscrm_required_plugins_notices()
{
    if (!is_plugin_active('gravityforms/gravityforms.php')) {
        echo "<div class=\"notice notice-error is-dismissible\"><p><strong>Please install <a href='https://www.gravityforms.com/' target='_blank'>GravityForms plugin</a> to make use of Infusionsoft-GravityForms Integrator plugin!</strong></p><button type=\"button\" class=\"notice-dismiss\"><span class=\"screen-reader-text\">Dismiss this notice.</span></button></div>";
    }

    if (!is_plugin_active('infusionsoft-sdk/infusionsoft-sdk.php')) {
        echo "<div class=\"notice notice-error is-dismissible\"><p><strong>Please install <a href='https://pl.wordpress.org/plugins/infusionsoft-sdk/' target='_blank'>Infusionsoft SDK plugin</a> to make use of Infusionsoft-GravityForms Integrator plugin!</strong></p><button type=\"button\" class=\"notice-dismiss\"><span class=\"screen-reader-text\">Dismiss this notice.</span></button></div>";
    } else {    
   		\Infusionsoft_AppPool::addApp(new \Infusionsoft_App(get_option('infusionsoft_sdk_app_name') . '.infusionsoft.com', get_option('infusionsoft_sdk_api_key')));
   		if (! \Infusionsoft_DataService::ping()) {
    		echo "<div class=\"notice notice-error is-dismissible\"><p><strong>Please set your Infusionsoft SDK app name and API key on the <a href=\"" . admin_url( 'options-general.php?page=infusionsoft-sdk/infusionsoft-sdk.php' ) . "\">settings page.</a></strong> to make use of Infusionsoft-GravityForms Integrator plugin!</strong></p><button type=\"button\" class=\"notice-dismiss\"><span class=\"screen-reader-text\">Dismiss this notice.</span></button></div>";
    	}   
    }
}

add_action('admin_notices', 'systasisgfifscrm\systasisgfifscrm_required_plugins_notices');